export CallbackSet,
       CallbackTiming,
       FunctionCallback,
       SaveCallback,
       PlotCallback,
       TimeAlignedSaveCallback


import Base: Filesystem


#######################################################################
#                                Types                                #
#######################################################################


"""
    AbstractCallback

Supertype for use of callback interface.
The following methods must be implemented by a `Callback<:AbstractCallback`:
- `(cb::Callback)(u, t, iter)` where `u,t,iter` are the current
  statevector, timestep and iteration index. It should return `true`
  if it was triggered and `false` otherwise.
"""
abstract type AbstractCallback end

#######################################################################
#                           Implementations                           #
#######################################################################


#######################################################################
#                           CallbackTiming                            #
#######################################################################


"""
    CallbackTiming(; every_iteration=0, every_dt=0)

Object to time whether callbacks should be run after `every_iteration` iterations
and/or after `every_dt` (simulation) time has passed.
Use `every_iteration = every_dt = 0` to disable any timing.
See also: `istimed!`
"""
mutable struct CallbackTiming
  every_iteration::Int64 # update interval for iterations
  every_dt::Float64      # update interval for time
  next_iter::Int64       # next scheduled update wrt iteration
  next_t::Float64        # next scheduled update wrt time
  t::Float64             # iteration time
  iter::Int64            # iteration count
  force::Bool            # ignore any timing above

  function CallbackTiming(; every_iteration=0, every_dt=0, force::Bool=false)
    if force
      @assert every_iteration == 0
      @assert every_dt == 0
    else
      @assert every_iteration >= 0
      @assert every_dt >= 0
    end
    return new(every_iteration, every_dt, every_iteration, every_dt, 0, 0, force)
  end
end


function state(cbt::CallbackTiming)
  """
  CallbackTiming
    t                   = $(cbt.t)
    next_t              = $(cbt.next_t)
    every_dt            = $(cbt.every_dt)
    iter                = $(cbt.iter)
    next_iter           = $(cbt.next_iter)
    every_iteration     = $(cbt.every_iteration)
  """
end


Base.show(io::IO, cbt::CallbackTiming) = print(io, state(cbt))


"""
    istimed!(timing::CallbackTiming, t)

Check if callback `timing` should trigger at time `t`.
If `true`, `timing` is updated and calling `istimed!(timing, t)` again will yield `false`
unless `timing.force = true`.
"""
function istimed!(timing::CallbackTiming, t)
  @unpack every_iteration, every_dt = timing

  timing.force && return true

  # already passed times don't increment state
  (t <= timing.t && timing.iter > 0) && return false

  timing.iter += 1
  timing.t    = t

  should_we_update_wrt_time = every_dt > 0 && timing.next_t <= t
  if should_we_update_wrt_time
    timing.next_t     += every_dt
  end

  should_we_update_wrt_iter = every_iteration > 0 && timing.next_iter <= timing.iter
  if should_we_update_wrt_iter
    timing.next_iter  += every_iteration
  end

  should_we_update = should_we_update_wrt_time || should_we_update_wrt_iter

  return should_we_update
end


#######################################################################
#                             CallbackSet                             #
#######################################################################


"""
    CallbackSet(::AbstractCallback...)

Use to queue a set of callbacks for evaluation, e.g.
```julia
CallbackSet(SaveCallback(...), PlotCallback(...), ...)
```
Callbacks are called in the order they are queued.
"""
struct CallbackSet
  callbacks::Vector{AbstractCallback}
  CallbackSet(callbacks...) = new([ cb for cb in callbacks ])
end


function (cbs::CallbackSet)(u, t, i=0)
  anycalled = false
  for idx = 1:length(cbs.callbacks)
    cb = cbs.callbacks[idx]
    status = cb(u, t, i)::Bool
    anycalled = anycalled || status
  end
  return anycalled
end


Base.push!(cbs::CallbackSet, cb::AbstractCallback) = push!(cbs.callbacks, cb)
Base.append!(cbs::CallbackSet, cb::Vector{<:AbstractCallback}) = append!(cbs.callbacks, cb)
Base.length(cbs::CallbackSet) = length(cbs.callbacks)


function Base.show(io::IO, cbs::CallbackSet)
  for cb in cbs.callbacks
    print(io, cb)
  end
end


#######################################################################
#                          FunctionCallback                           #
#######################################################################


"""
    FunctionCallback(fn::Function, timing::CallbackTiming)

Make callback for calling `fn(u,t)::Bool` at times controlled by `timing`.
"""
struct FunctionCallback{T_Function<:Function} <: AbstractCallback
  func::T_Function
  timing::CallbackTiming
end


function (cb::FunctionCallback)(u, t, iter; force=false)
  @unpack func, timing = cb
  if force || cb.timing.force || istimed!(timing, t)
    func(u, t)
    return true
  end
  return false
end


function state(fcb::FunctionCallback)
  """
FunctionCallback
  func                = $(nameof(fcb.func))
  timing              =
$(indent(state(fcb.timing), level=3, fillwith="  "))"""
end


Base.show(io::IO, fcb::FunctionCallback) = print(io, state(fcb))


#######################################################################
#                            SaveCallback                             #
#######################################################################


"""
    SaveCallback(variables::Symbol..., C::Cache, mesh::Mesh, filename, timing::CallbackTiming)

Make callback for saving `variables...` from `C` to file `filename`.
"""
mutable struct SaveCallback <: AbstractCallback
  _file::HDF5.File
  fcb::FunctionCallback
  filename::String
  variablenames::Vector{Tuple{Symbol,Symbol}}

  function SaveCallback(variables::Vector{Symbol}, C::Cache, mesh::Mesh, filename,
      timing::CallbackTiming)

    # normalize and sanitize filename
    filename = Filesystem.abspath(filename)
    if filename[end-2:end] != ".h5"
      filename = "$(filename).h5"
    end
    dir = Filesystem.dirname(filename)
    !isdir(dir) && error("Save directory '$dir' does not exist")
    isfile(filename) && error("File '$filename' already exsists")

    # gather list of registered variables
    fields = propertynames(C)
    varnames = [ dg1d.names(getproperty(C, f)) for f in fields ]

    # check if requested output variables are registered
    variablenames = Tuple{Symbol,Symbol}[]
    for var in variables
      idx = findfirst(vars -> var in vars, varnames)
      @assert !isnothing(idx) "Variable '$var' not defined"
      push!(variablenames, (fields[idx], var))
    end

    # open a file
    file = h5open(filename, "w")
    flush(file)

    # output function, to be called by FunctionCallback
    group_idx = 0 # use internal iteration index to keep contiguous group name increments
    function savefn(u, t)
      # generate group for new time step
      group_idx += 1
      group = create_group(file, "$(group_idx)")
      group["t"] = t
      for (field, var) in variablenames
        name = "$(field).$(var)"
        group[name] = getproperty(getproperty(C, field), var)
      end
      # important: use flush to not loose data
      flush(file)
    end

    fcb = FunctionCallback(savefn, timing)

    # save mesh now
    if length(variablenames) > 0
      group = create_group(file, "mesh")
      group["mesh"] = mesh.x
    end

    scb = new(file, fcb, filename, variablenames)

    # utilize a finalizer to close the file upon struct destruction
    # this also requires SaveCallback to be mutable
    # one could also use a RefValue and keep it immutable, 
    # see https://discourse.julialang.org/t/finalizer-only-works-with-mutable-structs/28777/11?u=fatteneder
    function close_file(scb)
      @async begin
        flush(savecallback._file)
        close(savecallback._file)
      end
    end

    finalizer(close_file, scb)
  end
end


(scb::SaveCallback)(u, t, iter; force=false) = scb.fcb(u, t, iter, force=force)


function state(scb::SaveCallback)
  """
  SaveCallback
    filename            = $(scb.filename)
    variablenames       = $(scb.variablenames)
  """
end

Base.show(io::IO, scb::SaveCallback) = print(io, state(scb))


#######################################################################
#                            PlotCallback                             #
#######################################################################


"""
    PlotCallback(plotfn::Function, cbt::CallbackTiming; pause::Real=false)

Make callback object for plotting with `plotfn(u,t)` at times controlled by `cbt`.
Set `pause` to a positive number (in seconds) to insert a break after every plot,
or set it to `true` to halt program.
"""
struct PlotCallback <: AbstractCallback
  fcb::FunctionCallback
  pause::Union{Bool,Float64}

  function PlotCallback(plotfn::Function, cbt::CallbackTiming; pause::Real=false)
    fcb = FunctionCallback(plotfn, cbt)
    if !(pause isa Bool)
      @assert pause > 0 "'pause > 0' required"
      pause = Float64(pause)
    end
    return new(fcb, pause)
  end
end


function (pcb::PlotCallback)(u, t, iter)
  @unpack fcb, pause = pcb

  status = fcb(u, t, iter)

  if status
    if pause isa Bool && pause
      println()
      @info "Press any button to continue"
      readline()
    else
      sleep(pause)
    end
  end

  return status
end


function state(pcb::PlotCallback)
  """
  PlotCallback
    pause               = $(pcb.pause)
  """
end

Base.show(io::IO, pcb::PlotCallback) = print(io, state(pcb))


#######################################################################
#                        TimeAlignedSaveCallback                      #
#######################################################################


"""
  TimeAlignedSaveCallback(times::Vector{T}, timestep_fn::Function, scb::SaveCallback) where T<:Real

Callback for aligning adaptive time stepping `timestep_fn(u, t)` with desired
`times` and save data with `scb`.
"""
struct TimeAlignedSaveCallback <: AbstractCallback
  aligned_timestep_fn::Function
  fcb::FunctionCallback

  function TimeAlignedSaveCallback(times::Vector{T}, timestep_fn::Function, scb::SaveCallback) where T<:Real
    @assert length(times) > 0 "Require at least one time stamp"
    @assert all(times .> 0) && all(diff(times) .> 0) """
    Time stamps must be all positive and increasingly ordered
    """
    @assert length(methods(timestep_fn, [Any,Real,Int])) > 0 """
    `timestep_fn` must take arguments `(u,t,n_attempts)`.
    """

    n_times = length(times)
    index_next_t = 1
    schedule_save = false
    # we wrap `timestep_fn` to adjust proposed step in case it exceeds the next desired time stamp
    function aligned_timestep_fn(u, t, n_attempts)
      dt = timestep_fn(u, t, n_attempts)
      if index_next_t <= n_times
        next_t = times[index_next_t]
        if t+dt > next_t
          dt = next_t - t # align time step
          index_next_t += 1
          schedule_save = true # trigger save through save_on_time below
        end
      end
      return dt
    end

    # use a FunctionCallback with this method to save the result of the aligned time stamp
    # we need to split the logic between these two functions, because the proposed
    # time step is computed before the time step is taken after which saving should happen
    function save_on_time(u, t; force=false)
      if schedule_save
        scb(u, t, 0; force=true)
        schedule_save = false
      end
    end
    fcb = FunctionCallback(save_on_time, CallbackTiming(every_iteration=1))

    return new(aligned_timestep_fn, fcb)
  end
end


(ftscb::TimeAlignedSaveCallback)(u, t, iter) = ftscb.fcb(u, t, iter)


function state(tascb::TimeAlignedSaveCallback)
  """
  TimeAlignedSaveCallback
    aligned_timestep_fn = $(nameof(tascb.aligned_timestep_fn))
    fcb                 = $(state(tascb.fcb))
  """
end

Base.show(io::IO, tascb::TimeAlignedSaveCallback) = print(io, state(tascb))
