export NumericalFluxes


#######################################################################
#                                Types                                #
#######################################################################


@doc """
  NumericalFluxes{TVar<:AbstractVariables}

Object for storing numerical fluxes appearing on `lhs, rhs` element faces.
"""
struct NumericalFluxes
  lhs::Variables
  rhs::Variables
  mesh::Mesh
end


#######################################################################
#                            Constructors                             #
#######################################################################


function NumericalFluxes(mesh::Mesh, names::Symbol...)
  _, ncells = layout(mesh)
  lhs, rhs = [ Variables(ncells, names...) for _ in 1:2 ]
  return NumericalFluxes(lhs, rhs, mesh)
end


#######################################################################
#                              Constants                              #
#######################################################################


@doc """
Outward pointing unit normal vectors for lhs/rhs edges of 1d cells, e.g.

              <--x nv_lhs           x--> nv_rhs
-----|-----|-----|-----|-----|------|-----|-----|-----|-----> x
"""
const nv_lhs = -1.0;
const nv_rhs =  1.0;


#######################################################################
#                              Deprecacted                              #
#######################################################################


# @doc """
#   flux_lax_friedrich!(numfluxes::NumericalFluxes{T}, fluxes::T, conservatives::T, speeds::T)
#
# Compute Lax Friedrich numerical flux inplace.
# """
# function numericalflux_lax_friedrich!(numfluxes::NumericalFluxes{T},
#     fluxes::T, conservatives::T, speeds::T) where T<:AbstractVariables
#   @unpack rhs, lhs, mesh = numfluxes
#   shape = layout(mesh)
#   npts, ncells = shape
#   vars = fieldnames(T)
#   for var in vars
#     mat_flux = reshape(getfield(fluxes, var), shape)
#     mat_cons = reshape(getfield(conservatives, var), shape)
#     mat_speed = reshape(getfield(speeds, var), shape)
#     numflux_lhs, numflux_rhs = getfield(lhs, var), getfield(rhs, var)
#     for idx in 1:ncells
#       idx_lhs, idx_rhs = [ dg1d.periodic_index(i, ncells) for i in [ idx-1, idx ] ]
#       f_lhs, f_rhs = mat_flux[end,idx_lhs], mat_flux[1,idx_rhs]
#       u_lhs, u_rhs = mat_cons[end,idx_lhs], mat_cons[1,idx_rhs]
#       C = maximum(mat_speed[:,idx])
#       avg_f = (f_lhs + f_rhs) / 2
#       dif_u = u_lhs - u_rhs
#       nf_lhs, nf_rhs = avg_f + C * dif_u * nv_lhs, avg_f + C * dif_u * nv_rhs
#       numflux_lhs[idx], numflux_rhs[idx] = nf_lhs, nf_rhs
#     end
#   end
#   return
# end


#######################################################################
#                             Deprecated                              #
#######################################################################


@doc """
Collect interface values, lhs/rhs refer to the lhs/rhs of the interface.
"""
struct Interface
  lhs
  rhs
end


@doc """
Evaluate fun with arguments intf and return result in a new Interface struct.
"""
function Interface(fun, intf::Interface...)
  intf_lhs = fun.(map(i->i.lhs, collect(intf))...)
  intf_rhs = fun.(map(i->i.rhs, collect(intf))...)
  return Interface(intf_lhs, intf_rhs)
end


@doc """
Project u onto all interfaces of mesh.
Periodic bdry conditions are assumed to set the ghost points.
"""
function to_interface(u, mesh::Mesh)
  @unpack element, K = mesh
  @unpack l_lhs, l_rhs, quadr = element
  F = K + 1   # number of interfaces

  lhs = similar(u, F)
  rhs = similar(u, F)
  if (quadr == :LGL || quadr == :LGL_lumped || quadr == :GLGL)
    # no interpolation needed
    lhs[2:end]    .= u[end,:]
    rhs[1:end-1]  .= u[1,:]
  else
    error("We don't handle grids without boundary points yet!")
    # lhs[2:end]    .= transpose(transpose(l_rhs) * u)
    # rhs[1:end-1]  .= transpose(transpose(l_lhs) * u)
  end # if
  # assume periodic grid
  lhs[1]        = lhs[end]
  rhs[end]      = rhs[1]

  intf = Interface(lhs, rhs)
  return intf
end


@doc """
Compute average between lhs and rhs interface values.
"""
function average(intf::Interface)
  return 0.5 * (intf.lhs + intf.rhs)
end


@doc """
Compute difference between lhs and rhs interface values.
"""
function difference(intf::Interface)
  return intf.lhs - intf.rhs
end


@doc """
Lax-Friedrich flux, see Hesthaven 2007, section 2.2.
C ... parameter to control upwinding contribution
      C = 0 turns this into central flux
      Common choices include absolute maxmium wave speed on global or
      local grid.
"""
function lax_friedrich_flux(avg_f, dif_u, normal, C)
  return avg_f + 0.5 * C * dif_u * normal
end


@doc """
Compute Lax-Friedrich flux times normal vector evaluated at lhs/rhs ends
of the subdomains.
C ... parameter to control upwinding contribution
      C = 0 turns this into central flux
      Common choices include absolute maxmium wave speed on global or
      local grid.
Flux is such that the conservation identity holds, 
cf. PhD thesis Zingan 2012, p. 77
     F(u_m, u_p, nv_m) = F(u_p, u_m, nv_p)
"""
function lax_friedrich_flux(intf_f::Interface, intf_u::Interface, C)
  avg_f = average(intf_f)
  dif_u = difference(intf_u)
  nf = lax_friedrich_flux.(avg_f, dif_u, nv_rhs, C)
  nf_lhs = nf[1:end-1]
  nf_rhs = nf[2:end]
  return nf_lhs, nf_rhs
end


@doc """
Special case of lax_friedrich_flux with C = 0.
Flux is such that the conservation identity holds, 
cf. PhD thesis Zingan 2012, p. 77
     F(u_m, u_p, nv_m) = F(u_p, u_m, nv_p)
"""
function central_flux(intf_u::Interface)
  nu = @. 0.5 * ( intf_u.lhs + intf_u.rhs )
  nu_lhs = nu[1:end-1]
  nu_rhs = nu[2:end]
  return nu_lhs, nu_rhs
end


@doc """
Flux based on interior values.
Useful for local DG method.
"""
function inner_flux(intf::Interface)
  nf_lhs = intf.rhs[1:end-1]
  nf_rhs = intf.lhs[2:end]
  return nf_lhs, nf_rhs
end


@doc """
Flux based on exterior values.
Useful for local DG method.
"""
function outer_flux(intf::Interface)
  nf_lhs = intf.lhs[1:end-1]
  nf_rhs = intf.rhs[2:end]
  return nf_lhs, nf_rhs
end


@doc """
Local DG flux according to Hesthaven, Warburton 2007, p. 252.
"""
function ldg_flux_asym_lhs(intf::Interface)
  nf_lhs = intf.lhs[1:end-1]
  nf_rhs = intf.lhs[2:end]
  return nf_lhs, nf_rhs
end


@doc """
Local DG flux according to Hesthaven, Warburton 2007, p. 252.
"""
function ldg_flux_asym_rhs(intf::Interface)
  nf_lhs = intf.rhs[1:end-1]
  nf_rhs = intf.rhs[2:end]
  return nf_lhs, nf_rhs
end


@doc """
  um ... cell interior value of u
  up ... cell exterior value of u
  nm ... cell outward pointing unit vector (at this bdry)
  β  ... 'upwinding direction'
"""
function ldg_flux(um, up, nm, β)
  avg_u = 0.5 * (um + up)
  dif_u = um - up
  return avg_u + nm * β * dif_u
end


@doc """
Symmetrized local DG flux according to Hesthaven, Warburton 2007, p. 252
Reduces to local DG rhs/lhs versions from above if β = ±1/2.
"""
function ldg_flux_sym(intf::Interface, β)
  nf_lhs =  ldg_flux.(intf.rhs[1:end-1], intf.lhs[1:end-1], nv_lhs, β)
  nf_rhs =  ldg_flux.(intf.lhs[2:end], intf.rhs[2:end], nv_rhs, β)
  return nf_lhs, nf_rhs
end


@doc """
  qm ... cell interior value of q
  qp ... cell exterior value of q
  um ... cell interior value of u
  up ... cell exterior value of u
  nm ... cell outward pointing unit vector (at this bdry)
  β  ... 'upwinding direction'
"""
function ldg_flux_penalty(qm, qp, um, up, nm, β, τ)
  avg_q = 0.5 * (qm + qp)
  dif_q = qm - qp
  dif_u = um - up
  return avg_q - nm * β * dif_q - nm * τ * dif_u
end


@doc """
Symmetrized and penalized local DG flux according to Hesthaven, 
Warburton 2007, p. 267
"""
function ldg_flux_sym_penalty( intf_q::Interface, intf_u::Interface, β, τ)
  nf_lhs =  ldg_flux_penalty.(
    intf_q.rhs[1:end-1], intf_q.lhs[1:end-1], 
    intf_u.rhs[1:end-1], intf_u.lhs[1:end-1],
    nv_lhs, β, τ)
  nf_rhs =  ldg_flux_penalty.(
    intf_q.lhs[2:end], intf_q.rhs[2:end], 
    intf_u.lhs[2:end], intf_u.rhs[2:end],
    nv_rhs, β, τ)
  return nf_lhs, nf_rhs
end


@doc """
Compute boundary flux from dirichlet boundary condition at given bdry.
"""
function bdry_flux(u, t, bdry::BDRY, bc::dg1d.Dirichlet_BC)
  return bc(t)
end


@doc """
Compute boundary flux from outflow boundary condition at given bdry.
"""
function bdry_flux(u, t, bdry::BDRY, bc::dg1d.Outflow_BC)
  bdryflux = 0.0
  if bdry == LHS
    bdryflux = u[1]
  else
    bdryflux = u[end]
  end
  return bdryflux
end


@doc """
Compute fluxes for given boundary conditions at both boundaries.
"""
function bdry_flux(u, t, bdryconds::dg1d.BdryConditions)
  bflux_lhs = bdry_flux(u, t, LHS, bdryconds.lhs)
  bflux_rhs = bdry_flux(u, t, RHS, bdryconds.rhs)
  return bflux_lhs, bflux_rhs
end


@doc """
Compute fluxes for LDG variable for given boundary conditions.

TODO Dispatch this on bc if we ever happen to need vonNeumann conditions.
"""
function bdry_flux_LDG_variable(q, t, bc::dg1d.Abstract_BdryConditions)
  bdryflux_lhs = 0.0
  bdryflux_rhs = 0.0
  return bdryflux_lhs, bdryflux_rhs
end
