@doc """
Lagrange interpolation l_i(x, {z}) polynomials.
Kopriva 2009, p. 74, (3.32)
m ... polynomial idx
x ... point on which to evaluate Lagrange polynomial
z ... grid points on which Lagrange polynomials are defined, l_i(z_j) = delta_ij
"""
function lagrange_poly(m, x, z)
  result = 1.0
  xm = z[m]
  for j in 1:length(z)
    (j == m) && continue
    xj = z[j]
    result *= (x - xj) / (xm - xj)
  end # for
  return result
end # function



@doc """
  `lagrange_interp_matrix(x, z)`

Interpolate data from reference grid `z` onto sample grid `x` by matrix multiplication.

TODO Rename this to interpolation matrix.
"""
function lagrange_interp_matrix(x, z)
  N = length(x)
  M = length(z)
  T = Matrix{Float64}(undef, N, M)
  for colIdx in 1:M
      T[:,colIdx] .= map(xi -> lagrange_poly(colIdx, xi, z), x)
  end # for
  return T
end # function



@doc """
Needed for derivative matrices for Lagrange polynomials.
Kopriva 2009, p. 74, (3.34).
x ... Interpolation nodes.
"""
function barycentric_weights(x)
  N = length(x)

  w = zeros(N)
  for i = 1:N
    xi = x[i]
    p = 1.0
    for j = 1:N
      (i == j) && continue
      xj = x[j]
      p *= (xj - xi)
    end
    w[i] = 1.0 / p
  end

  return w
end



@doc """
Kopriva 2009, p. 74, (3.36).
"""
function barycentric_interp(x, z, w, f)
  idx = findfirst(zk->abs(zk - x) ≈ 0, z)
  !isnothing(idx) && return f[idx]
  t = @. w / (x - z)
  num = dot(t, f)
  denom = sum(t)
  return num / denom
end


@doc """
Kopriva 2009, p. 75, (3.38).
"""
function barycentric_interp_matrix(x, z)
  w = barycentric_weights(z)
  Nx = length(x)
  Nz = length(z)
  @assert length(w) == Nz
  A = zeros(Nx, Nz)
  for k = 1:Nx
    xk = x[k]
    idxmatch = findfirst(zk->abs(zk - xk) ≈ 0, z)
    hasmatch = !isnothing(idxmatch)
    if hasmatch
      A[k, idxmatch] = 1
    else
      @. A[k,:] = w / (xk - z)
      A[k,:] ./= sum(A[k,:])
    end
  end

  return A
end



@doc """
Enhance accuracy of derivativ matrix D by computing diagonal elements such
that derivative of a constant function vanishes.
Kopriva 2009, p. 55, (2.43).
"""
function negative_sum_trick!(D)
  N,M = size(D)
  @assert N == M
  for i = 1:N
    D[i,i] = 0.0
    row = D[i,:]
    sortidx = sortperm(abs.(row)) # sorts ascending
    sortrow = row[sortidx]
    # manually sum in ascending order
    for j = 1:N
      D[i,i] -= sortrow[j]
    end
  end
end



"""
    first_derivative_matrix(n, x)

Compute `1`st nodal derivative matrix based on point set `x`.
Taken from Kopriva 2009, p. 82, (3.48).
"""
function first_derivative_matrix(x)

  N = length(x)
  bw = barycentric_weights(x)
  D = zeros(N, N)

  for i = 1:N, j = 1:N
    xi = x[i]; xj = x[j]
    bwi = bw[i]; bwj = bw[j]
    D[i,j] = bwj / (bwi * (xi - xj))
  end

  negative_sum_trick!(D)

  return D
end


"""
    nth_derivative_matrix(n, x)

Compute `n`th nodal derivative matrix based on point set `x`.
Taken from Kopriva 2009, p. 82, (3.50).
"""
function nth_derivative_matrix(n, x)

  N = length(x)
  bw = barycentric_weights(x)
  Dnm1 = derivative_matrix(n-1, x)
  Dn = zeros(Float64, size(Dnm1))

  for i = 1:N, j = 1:N
    xi = x[i]; xj = x[j]
    bwi = bw[i]; bwj = bw[j]
    Dn[i,j] = n / (xi - xj) * (bwj / bwi * Dnm1[i,i] - Dnm1[i,j])
  end

  negative_sum_trick!(Dn)

  return Dn
end


function derivative_matrix(k, x)
  @assert k > 0 "Invalid derivative order `k = $k` requested!"
  Dk = if k == 1
    first_derivative_matrix(x)
  else
    nth_derivative_matrix(k, x)
  end
  return Dk
end



"""
    vandermonde_matrix_legendre(z)

Establishes connection between nodal and modal coefficients for the Legendre basis, based on
the assumption that `u ≈ u_nodal(z_i) = u_modal(z_i)`.
`z` is the nodal point set at which the approximations should agree on.
"""
function vandermonde_matrix_legendre(z)
  Npts = length(z)
  V = [ legendre(z[i], j-1) / sqrt(2 / (2*(j-1)+1)) for i=1:Npts, j=1:Npts ]
  return V
end


"""
    vandermonde_matrix(basis_fn::Function, z)

Establishes connection between nodal and modal coefficients for a specific basis, based on
the assumption that `u ≈ u_nodal(z_i) = u_modal(z_i)`.
`basis_fn(z[i], j)` must return the function value of the `j`th modal basis function
evaluated at node `z[i]`, where the `N+1` basis functions are labeled as `j = 0,...,N`.
`z` is the nodal point set at which the approximations should agree on.
"""
function vandermonde_matrix(basis_fn::Function, z)
  Npts = length(z)
  V = [ basis_fn(z[i], j-1) for i=1:Npts, j=1:Npts ]
  return V
end


@deprecate vandermonde_matrix(z) vandermonde_matrix_legendre(z)


@doc """
Spectral filtering matrix, motivated by an artificial viscosity regularization.
Cf. Hesthaven, Warburton 2007, section 5.3,
PhD thesis Glaubitz 2020, section 8.5.
see hesthaven 2008, section 5.3
"""
function filter_matrix(z, α = 36.0, Nc = 0, s = 16.0)
  Npts = length(z)
  N = Npts - 1
  σ(n) = exp(- α * ((n - Nc) / (N - Nc))^s)
  diagL = ones(Npts)
  diagL[Nc+1:Npts] = [ σ(i) for i=Nc+1:Npts ]
  L = diagm(diagL)
  return L
end
