"""
    Polytrope <: AbstractEquationOfState

Polytropic equation of state

  p = K ρ^Γ

- `p` pressure
- `ρ` polytropic constant
- `K` polytropic constant
- `Γ` polytropic constant
"""
Base.@kwdef struct Polytrope <: ColdEquationOfState
  K::Float64
  Gamma::Float64
end


#######################################################################
#                             Parameters                              #
#######################################################################


function dg1d.parse_parameters(::Type{Polytrope}, prms=Dict())
  K = query(Float64, prms, "K", default=100.0)
  Gamma = query(Float64, prms, "Gamma", default=2.0)
  @assert K > 0     "Polytropic constant must be positive!"
  @assert Gamma > 0 "Polytropic index must be positiv!"
  return (; K, Gamma)
end


#######################################################################
#                            Computations                             #
#######################################################################


# p = p(ρ,ϵ)
function (eos::Polytrope)(::Type{Pressure},
                          ::Type{Density}, rho)
  @unpack K, Gamma = eos
  return K * rho^Gamma
end


# ρ = ρ(ϵ,p)
function (eos::Polytrope)(::Type{Density},
                          ::Type{Pressure}, p)
  @unpack K, Gamma = eos
  return (p / K)^(1/Gamma)
end


# h = h(ρ,p)
function (eos::Polytrope)(::Type{Enthalpy},
                          ::Type{Density}, rho)
  @unpack K, Gamma = eos
  return 1 + Gamma / (Gamma - 1) * K * rho^(Gamma - 1)
end


# ϵ = ϵ(ρ,p)
function (eos::Polytrope)(::Type{InternalEnergy},
                          ::Type{Density}, rho)
  @unpack K, Gamma = eos
  return K / (Gamma - 1) * rho^(Gamma - 1)
end


function (eos::Polytrope)(::Type{InternalEnergy},
                          ::Type{Pressure}, p)
  @unpack K, Gamma = eos
  return K / (Gamma - 1) * (p / K)^((Gamma-1)/Gamma)
end


# ∂p/∂ρ(ρ,ϵ)
function (eos::Polytrope)(::Type{Derivative{Pressure, Density}},
                          ::Type{Density}, rho)
  @unpack K, Gamma = eos
  return K * Gamma * rho^(Gamma-1)
end


# ∂p/∂ϵ(ρ,ϵ)
function (eos::Polytrope)(::Type{Derivative{Pressure, InternalEnergy}},
                          ::Type{Density}, rho)
  return 0
end
