function plot(rhs::RHS)
  plt_eq   = plot_equation(rhs)
  plt_rhs  = nothing # plot_rhs(rhs)
  plt_hrsc = plot_hrsc(rhs, rhs.hrsc)
  plt_tci  = plot_tci(rhs, rhs.tci)

  # drop missing plots and figure out layout
  list_sub_plts = [ plt_eq, plt_rhs, plt_hrsc, plt_tci ]
  filter!(!isnothing, list_sub_plts)
  layout = (length(list_sub_plts), 1)
  sub_plt = Plots.plot(list_sub_plts..., layout=layout)

  # combine plots
  @unpack cache = rhs
  @unpack t = get_global_variables(cache)
  plt = Plots.plot(sub_plt, layout=layout, size=(1400,1000), plot_title="t=$(t[1])",
                   left_margin=10Plots.mm)
  display(plt)
end


#######################################################################
#                              Equation                               #
#######################################################################


function plot_equation(rhs)

  @unpack cache, mesh = rhs
  @unpack rho, q, E   = get_dynamic_variables(cache)
  @unpack p, eps, S   = get_static_variables(cache)
  @unpack t           = get_global_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  L       = layout(mesh)
  _, K    = L

  rho = reshape(rho, L)
  q   = reshape(q, L)
  p   = reshape(p, L)
  E   = reshape(E, L)
  S   = reshape(S, L)
  eps = reshape(eps, L)
  x   = reshape(x, L)

  vars    = [ rho, q, E, p, eps, S ]
  ylabels = [ "rho", "q", "E", "p", "eps", "S" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    isnothing(var) && continue
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(3,2))
 
  return plt
end


#######################################################################
#                                HRSC                                 #
#######################################################################


function plot_hrsc(rhs, hrsc::Nothing) end


function plot_hrsc(rhs, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache, mesh = rhs
  @unpack mu          = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "μ"
  L       = layout(mesh)
  _, K = L

  mu = ones(L) .* mu'
  x  = reshape(x, L)

  plt = Plots.plot(xlabel=xlabel, ylabel=ylabel, title="Artificial viscosity")
  for k in 1:K
    Plots.plot!(plt, x[:,k], mu[:,k], label=false)
  end

  return plt
end


function plot_hrsc(rhs, hrsc::HRSC.SmoothedArtificialViscosity)

  plt_mu = plot_hrsc(rhs, hrsc.av)

  @unpack cache, mesh = rhs
  @unpack smoothed_mu = get_static_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "μ"
  L       = layout(mesh)
  _, K    = L

  smoothed_mu = reshape(smoothed_mu, L)
  x           = reshape(x, L)

  plt_smoothed_mu = Plots.plot(xlabel=xlabel, ylabel=ylabel,
                               title="Smoothed Artificial Viscosity")
  for k in 1:K
    Plots.plot!(plt_smoothed_mu, x[:,k], smoothed_mu[:,k], label=false)
  end

  combined_plt = Plots.plot(plt_mu, plt_smoothed_mu, layout=(1,2))

  return combined_plt
end


plot_hrsc(rhs, hrsc::AbstractReconstruction) = nothing


#######################################################################
#                                 TCI                                 #
#######################################################################


function plot_tci(rhs, tci::Nothing) end


function plot_tci(rhs, tci::TCI.AbstractTCI) 

  @unpack cache, mesh = rhs
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylimits = (-0.5,1.5)
  ylabel  = "Entropy production"
  L       = layout(mesh)
  _, K    = L

  flag = ones(L) .* flag'
  x    = reshape(x, L)

  vars    = [ flag ]
  ylabels = [ "flag" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i],
                         xlimits=xlimits, ylimits=ylimits) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt
end


function plot_tci(rhs, tci::TCI.EntropyProduction)

  @unpack cache, mesh = rhs
  @unpack EP          = get_static_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "Entropy production"
  L       = layout(mesh)
  _, K = L

  flag = ones(L) .* flag'
  EP   = reshape(EP, L)
  x    = reshape(x, L)

  vars    = [ flag, EP ]
  ylabels = [ "flag", "|∂t S + ∂x F_S|" ]
  titles  = [ "Troubled Cell Indicator", "Entropy Production" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i], title=titles[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt
end


#######################################################################
#                                 RHS                                 #
#######################################################################


function plot_rhs(rhs)

  @unpack cache, mesh = rhs
  @unpack rhs_rho, rhs_q, rhs_E = get_rhs_variables(cache)
  @unpack t                     = get_global_variables(cache)
  @unpack x                     = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  L       = layout(mesh)
  _, K    = L

  rhs_rho = reshape(rhs_rho, L)
  rhs_q   = reshape(rhs_q, L)
  rhs_E   = reshape(rhs_E, L)
  x   = reshape(x, L)

  vars    = [ rhs_rho, rhs_q, rhs_E, nothing ]
  ylabels = [ "rhs_rho", "rhs_q", "rhs_E", nothing ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    isnothing(var) && continue
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(2,2))
 
  return plt
end
