Base.@kwdef struct RHS{T_RSolver  <:AbstractRiemannSolver,
                       T_HRSC     <:Maybe{HRSC.AbstractHRSC},
                       T_TCI      <:Maybe{TCI.AbstractTCI},
                       T_LDG_RSolver<:Maybe{AbstractRiemannSolver},
                       T_AV_RSolver<:Maybe{AbstractRiemannSolver}}

  cache::Cache
  mesh::Mesh
  equation::Equation
  rsolver::T_RSolver
  hrsc::T_HRSC               = nothing
  tci::T_TCI                 = nothing
  ldg_rsolver::T_LDG_RSolver = nothing
  av_rsolver::T_AV_RSolver   = nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(rhs::RHS)
  register_variables!(rhs.cache,
    dynamic_variablenames = (:sD, :sSr, :stau,),            # 'scaled' conservatives
    rhs_variablenames     = (:rhs_sD, :rhs_sSr, :rhs_stau), # rhs of conservatives
    static_variablenames  = (:flx_sD, :flx_sSr, :flx_stau,  # conservatives' fluxes
                             :src_sD, :src_sSr, :src_stau,  # sources
                             :D, :Sr, :tau,                 # true hydrodynamic conservatives
                             :rho, :vr, :p, :eps,        # primitives
                             :max_v, :r,                 # maximum charactersitic speed, grid
                             :E, :Em1,                   # entropy
                             :flx_E, :flx_Em1,           # entropy flux
                             :A, :B, :A_r, :B_r),        # metric fields and derivatives
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  _register_variables!(rhs, rhs.hrsc)
  _register_variables!(rhs, rhs.tci)
  @unpack r = get_static_variables(rhs.cache)
  @. r = rhs.mesh.x[:]
end


function _register_variables!(rhs, tci_or_hrsc::Nothing) end


_register_variables!(rhs, hrsc::AbstractHRSC) = 
  error("_register_variables: Not implemented for hrsc of type '$(typeof(hrsc))'")

_register_variables!(rhs, hrsc::HRSC.AbstractReconstruction) = nothing


function _register_variables!(rhs, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(rhs.cache,
    static_variablenames  = (:ldg_sD, :ldg_sSr, :ldg_stau,              # local DG variable
                             :flx_ldg_sD, :flx_ldg_sSr, :flx_ldg_stau), # local DG flux
    cell_variablenames    = (:mu,), # 'one viscosity to rule them all'
    )
end


function _register_variables!(rhs, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(rhs, hrsc.av)
  register_variables!(rhs.cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(rhs, tci::TCI.AbstractTCI)
  register_variables!(rhs.cache, cell_variablenames = (:flag,:sD_flag,:sSr_flag,:stau_flag))
end


function _register_variables!(rhs, tci::TCI.EntropyProduction)
  register_variables!(rhs.cache, 
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end


#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(state_u, t, rhs::RHS)
  wrap_dynamic_variables!(rhs.cache, state_u)
  return timestep(rhs, rhs.hrsc)
end


function timestep(rhs::RHS, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh, equation = rhs
  @unpack max_v = get_static_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N^2 * vmax)
  return dt
end


function timestep(rhs::RHS, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  max_v = get_variable(cache, :max_v)
  vmax = dg1d.absolute_maximum(view(max_v,:))
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(state_du, state_u, t, rhs::RHS, bdryconds::BoundaryConditions,
    ldg_bdryconds::Maybe{BoundaryConditions}, av_bdryconds::Maybe{BoundaryConditions})

  @unpack cache = rhs
  wrap_dynamic_variables!(cache, state_u)
  wrap_rhs_variables!(cache, state_du)
  rhs!(rhs, t, rhs.hrsc, bdryconds, ldg_bdryconds, av_bdryconds)
end


function rhs!(rhs::RHS, t, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack sD, sSr, stau             = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau = get_bdry_variables(cache)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  @unpack p, vr, rho, eps = get_static_variables(cache)

  compute_rhs_weak_form!(rhs_sD,   flx_sD,    lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,   lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau,  lhs_numflx_stau, rhs_numflx_stau, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractReconstruction,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack sD, sSr, stau                    = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau        = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau        = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau = get_bdry_variables(cache)
  @unpack flag                             = get_cell_variables(cache)

  HRSC.reconstruct!(sD,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(sSr,   flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(stau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  compute_rhs_weak_form!(rhs_sD,   flx_sD,    lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,   lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau,  lhs_numflx_stau, rhs_numflx_stau, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver,
          ldg_rsolver, av_rsolver                      = rhs
  @unpack sD, sSr, stau                          = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau,
          ldg_sD, ldg_sSr, ldg_stau,
          flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau  = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau              = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau,
          lhs_numflx_ldg_sD,   rhs_numflx_ldg_sD,
          lhs_numflx_ldg_sSr,  rhs_numflx_ldg_sSr,
          lhs_numflx_ldg_stau, rhs_numflx_ldg_stau = get_bdry_variables(cache)

  # broadcast_volume!(cons2prim, equation, cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, t)
  compute_rhs_weak_form!(ldg_sD,   flx_ldg_sD,   lhs_numflx_ldg_sD,   rhs_numflx_ldg_sD,   mesh)
  compute_rhs_weak_form!(ldg_sSr,  flx_ldg_sSr,  lhs_numflx_ldg_sSr,  rhs_numflx_ldg_sSr,  mesh)
  compute_rhs_weak_form!(ldg_stau, flx_ldg_stau, lhs_numflx_ldg_stau, rhs_numflx_ldg_stau, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, t)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  @. flx_sD          += flx_ldg_sD
  @. flx_sSr         += flx_ldg_sSr
  @. flx_stau        += flx_ldg_stau
  @. lhs_numflx_sD   += lhs_numflx_ldg_sD
  @. lhs_numflx_sSr  += lhs_numflx_ldg_sSr
  @. lhs_numflx_stau += lhs_numflx_ldg_stau
  @. rhs_numflx_sD   += rhs_numflx_ldg_sD
  @. rhs_numflx_sSr  += rhs_numflx_ldg_sSr
  @. rhs_numflx_stau += rhs_numflx_ldg_stau

  compute_rhs_weak_form!(rhs_sD,   flx_sD,   lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,  lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau, lhs_numflx_stau, rhs_numflx_stau, mesh)

  return true
end
