"""
    MDAViscosity

Modal Decay Averaged Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct MDAViscosity <: AbstractArtificialViscosity
  mesh::Mesh
  cmax::Float64 = 1.0
end


function parse_parameters(::Type{MDAViscosity}, prms)
  cmax = query(Float64, prms, "cmax", default=1.0)
  @assert cmax > 0 "'cmax > 0' required"
  return (;cmax)
end


function compute_viscosity!(
    mu, 
    u, v, 
    av::MDAViscosity)

  @unpack mesh = av
  @unpack cmax = av
  @unpack element, K, L, x = mesh
  @unpack invV, N = element

  # h = L / (K * N)
  h = L / K

  u = reshape(u, layout(mesh))
  uhat = invV * u  # modal coefficients

  # perfect decay modal coefficients, added to enhance sensitivity to 
  # floating point noise
  vhat = collect(Float64, 1:N).^(-N)
  vhat ./= norm(vhat)

  cellnorm2_u = dg1d.cellwise_inner_product(u, u, mesh)
  cellnorm2_u[abs.(cellnorm2_u) .< 1e-10 ] .= 1.0

  maxabs_v = extrema(abs.(v))[2]
  fill!(mu, cmax * h / N * maxabs_v)

  # estimate effective decay rate by least square fit of modal coefficients, (3.9)
  # sum_m ( log(|q_m|) - (log(c) - tau log(m)) )^2 = sum_m (y_m - (a + b x_m))^2 -> min
  # log(m) ... grid points in log scale
  rng_N = 1:N
  data_x = @. - log(rng_N)

  for j in 1:K
    # (3.11)
    # only consider coefficients 2:Np (which corresponds to 1:N), because
    # the 1st (0th) one corresponds to a constant mode that is irrelevant
    # for the smoothness esimation.
    mod_uh = @. sqrt( uhat[2:end,j]^2 + cellnorm2_u[j] * vhat^2 )
    # (3.12)
    skylined_mod_uh = [ maximum(mod_uh[min(n,N-1):end]) for n in 1:N ]

    # data points in log scale
    data_y = @. log(skylined_mod_uh)
    data_y[isinf.(data_y)] .= -20.0
    _, _, tau = dg1d.least_sqr_linear(data_x, data_y)
    isnan(tau) && error("undesired nan encountered")

    if 1.0 <= tau <= 3.0
      mu[j] *= 1.0 - 0.5 * (tau - 1.0)
    elseif tau > 3.0
      mu[j] *= 0.0
    end
  end

  return true
end
