function plot(rhs::RHS)
  plt_eq   = plot_equation(rhs, rhs.equation)
  plt_hrsc = plot_hrsc(rhs, rhs.hrsc)
  plt_tci  = plot_tci(rhs, rhs.tci)

  # drop missing plots and figure out layout
  plts = [ plt_eq, plt_hrsc, plt_tci ]
  filter!(!isnothing, plts)
  layout = (length(plts), 1)

  # combine plots
  @unpack cache = rhs
  @unpack t = get_global_variables(cache)
  plt = Plots.plot(plts..., layout=layout, size=(1400,1000), plot_title="t=$(t[1])",
                   left_margin=10Plots.mm)
  display(plt)
end


#######################################################################
#                              Equation                               #
#######################################################################


function plot_equation(rhs, eq::AbstractScalarEquation)

  @unpack cache, mesh = rhs
  @unpack u           = get_dynamic_variables(cache)
  @unpack x           = mesh

  xlimits = extrema(x)
  L       = layout(mesh)
  _, K = L

  u = reshape(u, L)
  x = reshape(x, L)

  xlabel  = "x"
  ylabel = "u" 
  plt    = Plots.plot(xlabel=xlabel, ylabel=ylabel)
  for k in 1:K
    Plots.plot!(plt, x[:,k], u[:,k], label=false)
  end

  return plt
end


#######################################################################
#                                HRSC                                 #
#######################################################################


function plot_hrsc(rhs, hrsc::Nothing) end


function plot_hrsc(rhs, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache, mesh = rhs
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "μ"
  L       = layout(mesh)
  _, K = L

  mu = if hrsc isa HRSC.SmoothedArtificialViscosity
    @unpack smoothed_mu = get_static_variables(cache)
    reshape(smoothed_mu, L)
  else
    @unpack mu = get_cell_variables(cache)
    mu = ones(L) .* mu'
  end
  x  = reshape(x, L)

  plt = Plots.plot(xlabel=xlabel, ylabel=ylabel)
  for k in 1:K
    Plots.plot!(plt, x[:,k], mu[:,k], label=false)
  end

  return plt
end


plot_hrsc(rhs, hrsc::Union{HRSC.BernsteinReconstruction,
                           HRSC.WENOReconstruction}) = nothing


#######################################################################
#                                 TCI                                 #
#######################################################################


function plot_tci(rhs, tci::Nothing) end


function plot_tci(rhs, tci::TCI.AbstractTCI) 

  @unpack cache, mesh = rhs
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  L       = layout(mesh)
  _, K    = L

  flag = ones(L) .* flag'
  x    = reshape(x, L)

  vars    = [ flag ]
  ylabels = [ "flag" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
    Plots.plot!(plt, ylimits=(0.0,1.0))
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt

end


function plot_tci(rhs, tci::TCI.EntropyProduction)

  @unpack cache, mesh = rhs
  @unpack EP          = get_static_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "Entropy production"
  L       = layout(mesh)
  _, K = L

  flag = ones(L) .* flag'
  EP   = reshape(EP, L)
  x    = reshape(x, L)

  vars    = [ flag, EP ]
  ylabels = [ "flag", "|∂t E + ∂x F|" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt

end
