Base.@kwdef struct ModalDecayHighest <: AbstractTCI
  mesh
  # parameters
  threshold_min::Float64
  threshold_max::Float64
end


function parse_parameters(::Type{ModalDecayHighest}, prms)
  threshold_min = query(Float64, prms, "threshold_min", default=1.0)
  threshold_max = query(Float64, prms, "threshold_max", default=3.0)
  @assert threshold_min > 0
  @assert threshold_max > 0
  @assert threshold_min < threshold_max
  return (; threshold_min, threshold_max)
end


function compute_indicator!(
    flag, # mutated outputs
    u,    # inputs,
    tci::ModalDecayHighest)

  @unpack mesh                 = tci
  @unpack invjac, dl           = mesh
  @unpack D, invV, V, iproduct = mesh.element

  L       = layout(mesh)
  Npts, K = L
  N       = Npts - 1
  u       = reshape(u, L)

  # perfect decay modal coefficients, added to enhance sensitivity to 
  # floating point noise
  vhat = collect(Float64, 1:N).^(-N)
  vhat ./= norm(vhat)
                                                                                                    
  diff_uk  = zeros(Npts)
  @unpack threshold_min, threshold_max = tci
  for k in 1:K

    uk                = u[:,k]
    diff_uk          .= invV * uk
    diff_uk[1:end-1] .= 0.0
    diff_uk          .= V * diff_uk
    norm2_diff_uk     = iproduct(diff_uk, diff_uk) / invjac[k]
    norm2_uk          = iproduct(uk, uk) / invjac[k]
                                                                                                    
    S = abs(norm2_diff_uk/norm2_uk)
    S = abs(norm2_uk) < 1e-12 ? 1e-20 : S
    S = abs(S) == 0 ? 1e-20 : S
                                                                                                    
    (isinf(S) || isnan(S)) && error("Encountered Inf or NaN in estimation of highest mode energy!")

    # sensor
    tau = log10(S)

    if threshold_max < tau
      flag[k] = 0.0
    elseif threshold_min <= tau <= threshold_max
      flag[k] = 1.0 - (tau - threshold_min) / (threshold_max - threshold_min)
    else
      flag[k] = 1.0
    end
  end
                                                                                                    
  return
end
