export Segment


#######################################################################
#                                Types                                #
#######################################################################


@doc """
  Segment

Object for restricting the view onto a mesh to certain regions.
Used for debugging and inspecting results.
"""
struct Segment
  mesh::Mesh
  pts::UnitRange
  cells::UnitRange
end


@doc """
  VariableSegment

Object to restrict variables onto a segment on the mesh.
Used for debugging and inspecting results.
Use a `Segment` to generate an instance, e.g.
```julia
segment = Segment(mesh, 1:4, 3:4) # restrict to points 1:4 and cells 3:4
display(segment(vars))
```

See also: `Segment`.
"""
struct VariableSegment
  vars::Variables
  segment::Segment
end


#######################################################################
#                            Constructors                             #
#######################################################################


@doc """
  Segment(mesh[, i0:i1, j0:j1])

Construct a segment of `mesh` specified through index ranges `i0:i1,j0:j1`,
where `i, j` refer to `i`th point and `j`th cell in `mesh`.
Use ':' as index to select all.
"""
Segment(mesh, i::Int, cells::UnitRange)  = Segment(mesh, i:i, cells)
Segment(mesh, pts::UnitRange, j::Int)    = Segment(mesh, pts, j:j)
Segment(mesh, i::Int, j::Int)            = Segment(mesh, i:i, j:j)
Segment(mesh, ::Colon, cells)            = Segment(mesh, 1:mesh.element.Npts, cells)
Segment(mesh, pts, ::Colon)              = Segment(mesh, pts, 1:mesh.K)
Segment(mesh, ::Colon, ::Colon)          = Segment(mesh, 1:mesh.element.Npts, 1:mesh.K)
Segment(mesh)                            = Segment(mesh, :, :)


function (segment::Segment)(vars::Variables)
  return VariableSegment(vars, segment)
end


#######################################################################
#                              Internals                              #
#######################################################################


const _cell_colors = [ :red, :green, :yellow, :blue, :magenta, :cyan ]


#######################################################################
#                               Methods                               #
#######################################################################


# pretty printing


function Base.show(io::IO, arraysegment::VariableSegment)
  @unpack segment, vars = arraysegment
  @unpack mesh, pts, cells = segment
  shape = layout(mesh)
  data_x = vec(mesh.x[pts,cells])
  data_f = hcat([ reshape(f, shape)[pts, cells][:] for f in fields(vars) ]...)
  data = hcat(data_x, data_f)
  header = [ "x", [ name for name in names(vars) ]... ]
  n = pts.stop - pts.start + 1
  highlighters = []
  formatters = ft_printf("%0.12e")
  for j_cell in eachindex(cells)
    fnctr = (data, i, j) -> (i >= (j_cell-1)*n+1) && (i <= j_cell * n)
    clr = Crayon(foreground=_cell_colors[periodic_index(j_cell, length(_cell_colors))])
    push!(highlighters, Highlighter(fnctr, clr))
  end
  n_rows = size(data)[1]
  row_names = [ "" for i in 1:n_rows  ]
  for (i,cell) in enumerate(cells)
    row_names[(i-1)*n+1] = "$cell"
  end
  pretty_table(data, 
               header=header, 
               highlighters=Tuple(highlighters),
               formatters=formatters, 
               # body_hlines=collect(n:n:n_rows-n),
               crop=:none,
               row_names=row_names,
               row_name_column_title="#")
end
