module TroubledCells


import dg1d: @with_kw, @unpack, Mesh


abstract type AbstractTroubledCellIndicator end


@with_kw struct MinmodIndicator <: AbstractTroubledCellIndicator
  "Threshold for troubled cell indication, the lower the more sensitive"
  M::Float64 = 1e-5
  @assert M > 0
end



function minmod(args::T...)::T where T <: Real
  signs = sign.(args)
  if sum(abs.(diff(collect(signs)))) != T(0)
    return signs[1] * minimum(abs.(args))
  else
    return T(0)
  end
end


function minmod_M(dx::T, M::T, args::T...)::T where T <: Real
  if abs(args[1]) <= M * dx^2
    return args[1]
  else
    return minmod(args...)
  end
end


@doc """
  `find_troubled_cells(indicator::MinmodIndicator, u::AbstractArray, u_averages::AbstractArray, mesh::Mesh)`

Use modified `minmod` function to scan data `u` with averages `u_averages` to detect irregularities
in gradients. Returns an array of indices to troubled cells.
"""
function find_troubled_cells(indicator::MinmodIndicator, 
    u::AbstractArray, u_averages::AbstractArray, mesh::Mesh)

  @unpack K, dl = mesh
  @unpack Npts = mesh.element
  @unpack M = indicator
  list_trbld_cells = Int64[]
  shape = (Npts, K)
  mat_u = reshape(u, shape)

  # skip bydr cells for now
  for k in 2:K-1 # loop over cells
    for i in 1:Npts # loop over pts in cell
      ui = u[i]
      uavg = u_averages[k]
      ubdry_lhs = uavg - mat_u[1,k]
      ubdry_rhs = mat_u[end,k] - uavg
      delta_lhs = uavg - u_averages[k-1]
      delta_rhs = u_averages[k+1] - uavg

      # troubled check, cf. (3.64) in Bugner 2017
      if !(minmod_M(dl, M, ubdry_lhs, delta_lhs, delta_rhs) ≈ ubdry_lhs) ||
         !(minmod_M(dl, M, ubdry_rhs, delta_lhs, delta_rhs) ≈ ubdry_rhs)
        push!(list_trbld_cells, k)
        break
      end
    end
  end

  return list_trbld_cells
end


@with_kw struct ArtificialViscosityIndicator <: AbstractTroubledCellIndicator
  "Threshold for troubled cell indication, the lower the more sensitive"
  threshold::Float64 = 0.1
  @assert threshold >= 0
end


@doc """
  `find_troubled_cells(indicator::ArtificialViscosityIndicator, u::AbstractArray, mesh::Mesh)`

Flag cell troubled if `u` exceeds indicator.threshold value.
Returns list of troubled cell indices.
"""
function find_troubled_cells(indicator::ArtificialViscosityIndicator,
  u::AbstractArray, mesh::Mesh)

  idxs = findall(x -> x >= indicator.threshold, u)
  cells = [ Tuple(idx)[2] for idx in idxs ]
  unique!(cells)

  return cells
end


end
