@variables DummyVariables var1 var2 var3


function benchmark_fncall_staticarray(
    f_lhs::T, f_rhs::T, u_lhs::T, u_rhs::T, speed::Float64) where T<:AbstractVariables
  nfs = zeros(dg1d.StaticArrays.MVector{n_variables(T)})
  for i in 1:n_variables(T)
    fl = f_lhs[i]
    fr = f_rhs[i]
    ul = u_lhs[i]
    ur = u_rhs[i]
    nfs[i] = (fl + fr) / 2 - speed / 2 * (ul - ur)
  end
  return T(nfs)
end


function benchmark_fncall_components(
    f_lhs::T, f_rhs::T, u_lhs::T, u_rhs::T, speed::Float64) where T<:AbstractVariables
  f1_l, f2_l, f3_l = f_lhs
  f1_r, f2_r, f3_r = f_rhs
  u1_l, u2_l, u3_l = u_lhs
  u1_r, u2_r, u3_r = u_rhs
  nf1 = (f1_l - f1_r) / 2 - speed / 2 * (u1_l - u1_r)
  nf2 = (f2_l - f2_r) / 2 - speed / 2 * (u2_l - u2_r)
  nf3 = (f3_l - f3_r) / 2 - speed / 2 * (u3_l - u3_r)
  return T(nf1, nf2, nf3)
end


function benchmark_iteration_rowaccessor!(numfluxes, cons, fluxes)
  T = eltype(numfluxes)
  # fields_cons = zip(fields(cons)...)
  # for (i, (v1, v2, v3)) in enumerate(fields_cons)
  for i in 1:n_variables(T)
    ci = cons[i]
    fi = fluxes[i]
    numfluxes[i] = benchmark_fncall_staticarray(fi, fi, ci, ci, 1.0)
  end
end


function benchmark_iteration_components!(numfluxes, cons, fluxes)
  T = eltype(numfluxes)
  for i in 1:n_variables(T)
    c1 = cons.var1[i]
    c2 = cons.var2[i]
    c3 = cons.var3[i]
    f1 = fluxes.var1[i]
    f2 = fluxes.var2[i]
    f3 = fluxes.var3[i]
    ci = T(c1, c2, c3)
    fi = T(f1, f2, f3)
    numfluxes[i] = benchmark_fncall_staticarray(fi, fi, ci, ci, 1.0)
  end
end


@testset "Variables" begin


@testset "Constructors" begin
  
  # type defined by @variables macro
  vars = DummyVariables(1,2,3)
  # using keywords
  vars = DummyVariables(var1=1, var3=3, var2=2)

  # StructArray version of type defined by @variables macro =: VariablesArray
  # empty fields
  vararr = DummyVariablesArray()
  # zeroed fields
  vararr = DummyVariablesArray(10)

end


@testset "Interface" begin

  vars = DummyVariables(1,2,3)

  # get number of variables
  @test n_variables(vars) == 3

  # obtain references to underlying fields
  # === checks that they are indeed the same instance (and not a copy)
  fs = fields(vars)
  @test fs[1] === vars.var1
  @test fs[2] === vars.var2
  @test fs[3] === vars.var3
  # health checks
  @test isinf(vars) == false
  @test isnan(vars) == false
  vars = DummyVariables(1,NaN,Inf)
  @test isinf(vars) == true
  @test isnan(vars) == true

  # StructArrays
  N = 10
  varsarray = DummyVariablesArray(N)

  # Base methods
  @test length(varsarray) == N
  @info "Testing (pretty) printing of variables ..."
  display(fields(varsarray))
  display(varsarray)
  # health checks
  @test any(isinf.(varsarray)) == false
  @test any(isnan.(varsarray)) == false
  varsarray.var1[5] = Inf
  varsarray.var2[5] = NaN
  @test any(isinf.(varsarray)) == true
  @test any(isnan.(varsarray)) == true

end


@testset "Unsafe wrap" begin

  data = randn(3*100)
  # alias a StructArray with given data, e.g. changin data changes vars
  vars = alias_components(DummyVariablesArray, data)

  # same as above but using typeof and new and more data
  data = randn(3*200)
  vars = alias_components(typeof(vars), data)

  # elements 1:200 must coincide
  @test all(vars.var1 .== data[1:200])

end


@testset "Benchmark function calls" begin

  Npts = 500
  cons, flux = [ randn(DummyVariables) for _ in 1:2 ]
  @info "Benchmarking function calls with AbstractVariables"
  @info "Function with temporary StaticArrays"
  b = @benchmark benchmark_fncall_staticarray($flux, $flux, $cons, $cons, 1.0)
  display(b)
  @info "Function with components typed out"
  b = @benchmark benchmark_fncall_components($flux, $flux, $cons, $cons, 1.0)
  display(b)

end


@testset "Benchmark iterations" begin

  Npts = 500
  cons, flux, numflux = [ DummyVariablesArray(Npts) for _ in 1:3 ]
  @info "Benchmarking iterations over AbstractVariable structarrays"
  @info "Loop using row accessors"
  b = @benchmark benchmark_iteration_rowaccessor!($numflux, $flux, $cons)
  display(b)
  @info "Loop with typed out components"
  b = @benchmark benchmark_iteration_components!($numflux, $flux, $cons)
  display(b)

end
   

end
